use Utf8Checker;
use Db::Connection;
use Storage::Storage;
use Logging;
#==============================================================
#
# common subroutines
#

sub printToLog {
  my ($msg, $noeol) = @_;
  if ($msg) {
	print $msg;
  }
  unless ($noeol) {
	print "\n";
  }
}

sub printToError {
  my ($msg) = @_;
  print STDERR "$msg\n";
}

#BEGINNOTINRAQ2

sub makeFileReader
{
	my ($filename, $doNotFail) = @_;
	local (*INPUT);

	unless (-T $filename) {

		if (defined($doNotFail)) {
			&printToLog("Error: file '$filename' is not a text file");
		} else {
			&printToError("Error: file '$filename' is not a text file");
		}

		return undef;
	}

	unless (open(INPUT,"<$filename")) {

		binmode INPUT;
		if (defined($doNotFail)) {
			&printToLog("Error: unable to open file '$filename' to read");
		} else {
			&printToError("Error: unable to open file '$filename' to read");
		}

		return undef;
	}

	my $input = *INPUT;

	my $this = {
		'READ' => sub {
				my $a = <$input>;
				return $a;
			},

		'CLOSE' => sub { close($input); }

	};

	return $this;
}

sub makeArrayReader {
  my (@arr) = @_;
  my $size= 0;
  my $count = scalar(@arr);

  my $this={'READ'=>sub {
			  if($count == $size) {
				return undef;
			  }

			  return $arr[$count++];
			},
			'CLOSE'=>sub {
			}
		   };
  return $this;
}

sub makeParser {
  my $reader;
  my ($this,$ptrRows,$keySep,$valSep,$valKeySep,$comment,
      $typeRows,$maxIndex,$rowIndex);
  {
	$reader = shift;
	unless(defined($reader)) {
	  return undef;
	}

    my %args=(@_);
    $keySep = $args{'KEYSEPARATOR'};
    $valSep = $args{'VALUESEPARATOR'};
    $valKeySep = $args{'KEYVALUESEPARATOR'};
  }
  $this={'READER'=>sub {
		   if(@_){
			 $reader=shift;
		   }
		   return $reader;
		 },
		 'FILE'=>sub {
		   if(@_){
			 $reader = makeFileReader(@_);
		   }
		   return undef;
		 },
		 'KEYSEPARATOR'=>sub{
		   if(@_){
			 $keySep=shift;
		   }
		   return $keySep;
		 },
		 'VALUESEPARATOR'=>sub{
		   if(@_){
			 $valSep=shift;
		   }
		   return $valSep;
		 },
		 'KEYVALUESEPARATOR'=>sub{
		   if(@_){
			 $valKeySep=shift;
		   }
		   return $valKeySep;
		 },
		 'COMMENT'=>sub{
		   if(@_){
			 $comment=shift;
		   }
		   return $comment;
		 },
		 'PARSE'=>sub{
		   my %args=(@_);

		   $keySep=$args{'KEYSEPARATOR'} if exists $args{'KEYSEPARATOR'};
		   $valSep=$args{'VALUESEPARATOR'} if exists $args{'VALUESEPARATOR'};
		   $valKeySep=$args{'KEYVALUESEPARATOR'} if exists $args{'KEYVALUESEPARATOR'};
		   $comment=$args{'COMMENT'} if exists $args{'COMMENT'};

		   $ptrRows = &parseFile('reader'=>$reader,'keysep'=>$keySep,
								 'valsep'=>$valSep,'keyvalsep'=>$valKeySep,
								 'comment'=>$comment);
		   $typeRows = ref($ptrRows);
		   if($typeRows=~/ARRAY/){
			 $rowIndex=-1;
			 $maxIndex=scalar(@{$ptrRows})-1;
		   }
		   return $ptrRows;
		 },
		 'ROW'=>sub{
		   if($typeRows=~/HASH/){
			 my $key=shift;
			 return $ptrRows->{$key};
		   }elsif($typeRows=~/ARRAY/){
			 if($rowIndex<$maxIndex){
			   $rowIndex++;
			   return $ptrRows->[$rowIndex];
			 }else{
			   return undef;
			 }
		   }else{
			 return undef;
		   }
		 },
		};

  return $this;
}

sub makeFileParser {
  my $filename = shift;
  return makeParser(makeFileReader($filename), @_);
}

sub makeSafeFileParser {
  my $filename = shift;
  return makeParser(makeFileReader($filename, 'doNotFail'), @_);
}

sub makeStringParser {
  my $content = shift;
  return makeParser(makeArrayReader($content), @_);
}

sub parseFile {
  my %arg = (@_);
  my $reader = $arg{'reader'};
  my $keySep = $arg{'keysep'};
  my $valSep = $arg{'valsep'};
  my $valKeySep = $arg{'keyvalsep'};
  my $comment = $arg{'comment'};

  my ($ret,$key,$value,$reKey,$reValue,@values,$keyBlank,$valBlank,
     $vkey,$vvalue,$reKeyValue,$reComment);

  if($keySep){
    $ret = {};
    if(ref($keySep)=~/regexp/i || ($keySep eq ' ')){
      $reKey = $keySep;
    }else{
      $reKey = qr/$keySep/;
    }
  }else{
    $ret = [];
  }
  if($valSep){
    if(ref($valSep)=~/regexp/i || ($valSep eq ' ')){
      $reValue = $valSep;
    }else{
      $reValue = qr/$valSep/;
    }
  }
  if($valKeySep){
    if(ref($valKeySep)=~/regexp/i || ($valKeySep eq ' ')){
      $reKeyValue = $valKeySep;
    }else{
      $reKeyValue = qr/$valKeySep/;
    }
  }
  if($comment){
    if(ref($comment)=~/regexp/i){
      $reComment = $comment;
    }else{
      $reComment = qr/$comment/;
    }
  }

  while($_ = $reader->{'READ'}->()){
    chomp;
    next unless $_;
    if($comment && /$reComment/){
	  next;
    }
    if ($keySep){
      ($key,$value)=split($reKey,$_,2);
      if($key){
		$key=~s/^\s+//;
		$key=~s/\s+$//;
		if($valSep){

		  if ($valKeySep){
			$ret->{$key}={};
			foreach $value (split($reValue,$value)){
			  ($vkey,$vvalue)=split($reKeyValue,$value);
			  $vkey=~s/^\s+//;
			  $vkey=~s/\s+$//;
			  $ret->{$key}->{$value}=$vvalue;
			}
		  }else{
			push @{$ret->{$key}},split($reValue,$value);
		  }
		}else{
		  $value=~s/^\s+//;
		  $value=~s/\s+$//;
		  $ret->{$key}=$value;
		}
      }
    }else{
      if($valSep){
		push @{$ret},[split($valSep,$_)];
      }else{
		push @{$ret},$_;
      }
    }
  }

  $reader->{'CLOSE'}->();

  return $ret;
}

#ENDNOTINRAQ2

sub makePasswordNode {
  my $value= shift;
  my $type = shift||'encrypted';

  my $item = &makeXmlNode('password');

  if ($value eq '') { $type = 'plain'; }
  else { if ($type ne 'plain') { $type = 'encrypted'; } }

  $item->{'CONTENT'}->($value);  ## html encode
  $item->{'ATTRIBUTE'}->('type',$type);
  return $item;
}

sub mangleName {
  my $name = shift;

  if($name =~ /^\d/) {
    $name = 'user' . $name;
  }
  $name =~ s/\./_/g;

  return $name;
}

sub makeSysUserNode {
  my ($name, $password, $type) = @_;

  unless(defined($name) && defined($password) && defined($type)) {
	printToError("makeSysUserNode: Username, password or password type is not set.");
  }

  my $sysUserNode = makeXmlNode('sysuser');
  $sysUserNode->{'ATTRIBUTE'}->('name', mangleName($name));
  $sysUserNode->{'ADDCHILD'}->(makePasswordNode($password, $type));

  return $sysUserNode;
}

# Returns home directory for the system account $user
sub getHomeDir {
	my ($user) = @_;
	my $home;
	my @pw = getpwnam($user);
	if (-1 != $#pw) {
		$home = $pw[7];
	}
	unless ($home) {
		# how it is by default
		$home = '/home/$account';
	}
	return $home;
}

# Returns password for system account $user
sub getSystemPassword {
	my ($user) = @_;
	my $password;
	my @pw = getpwnam($user);
	if (-1 != $#pw) {
		$password = $pw[1];
	}
	unless ($password) {
		$password = '';
	}
	return $password;
}

sub createUserNameFromDomain {
  my ($name) = @_;

  if ($name =~ /^\d/) {
	$name = 'u' . $name;
  }

  $name =~ s/\./_/g;
  $name =~ s/^www\.//g;
  $name =~ s/[aoeui]//g if length($name) > 16;
  $name = substr($name, 0, 16) if length($name) > 16;

  return $name;
}

my $compress = 1;

sub setCompress {
  ($compress) = @_;
}

#FIXME: HACK
my $workDir;

sub initStorage {
  my $compress = !(exists $ptrArgs->{'no-compress'});

  $workDir = `pwd`;
  chomp $workDir;
  $workDir .= "/";

  # 2Gb-1 is a reasonable default split size
  my $splitsize = (2**31) - 1;
  $splitsize = $ENV{'PLESKX_SPLIT_SIZE'} if exists $ENV{'PLESKX_SPLIT_SIZE'};

  if (exists $ptrArgs->{'output'}) {
    $storage = Storage::Storage::createMimeStorage($compress, $ptrArgs->{'output'}, $splitsize);
  } else {
    $storage = Storage::Storage::createFileStorage($compress, $workDir, $splitsize);
  }
}

#
# Compatibility wrapper
#

sub makeDumpDb {
  my ($dbName, $dbType, $dbUser, $dbPassword, $dbHost, $ptrDomainBin, $variables, $domainSocket, $pg_manage) = @_;

  return if exists $ptrArgs->{'no-content'} or exists $ptrArgs->{'configuration-only'};

  my %params;

  $params{'name'} = $dbName;
  $params{'user'} = $dbUser;
  $params{'password'} = $dbPassword;
  $params{'host'} = $dbHost if $dbHost;
  $params{'socket'} = $domainSocket if $domainSocket;
  $params{'preload_dirs'} = $variables if $variables;

  if ($dbType eq 'Pg') {
    $dbType = 'postgresql';
    $params{'name'} = 'template1';
    if(defined $pg_manage) { $params{'pg_manage'} = $pg_manage; } # for plesk 7, bug 86423
  }

  $params{'type'} = $dbType;

  # check db existence, bug 83616
  my $connection = Db::Connection::getConnection(%params);
  my $db_exist = $connection->execute_rownum($dbType eq 'postgresql' ? "SELECT 1 FROM pg_database WHERE datname = '$dbName'" : 'SELECT 1');
  $connection->finish();
  $connection->disconnect();

  $params{'name'} = $dbName;

  return $storage->addDb("$dbName.$dbType.sql", %params) if $db_exist;

  Logging::error("Database $dbName is not accessible for dumping");

  return undef;
}

sub addDbUsers {
  my ($dbNode, $dbName, $wrapUserMysql) = @_;

  printToLog("Dupming $dbName database users");

  $dbName =~ s/_/\\_/g;

  my (@users, $dbUser, $ptrRow);
  my $sql = "SELECT DISTINCT User FROM db WHERE Db LIKE '$dbName'";
  $wrapUserMysql->{'EXECUTE'}->($sql);
  while ($ptrRow = $wrapUserMysql->{'FETCHROW'}->()) {
		push @users, (@{$ptrRow})[0];
  }
  $wrapUserMysql->{'FINISH'}->();

  foreach $dbUser (@users) {
		addDbUser($dbNode, $dbUser, $wrapUserMysql);
  }
}

sub addDbUser {
  my ($dbNode, $user, $wrapUserMysql, $passwordNode, $accessHosts) = @_;

  &printToLog("Dumping mysql user '$user'");

	my ( $item );

  if ( ref( $wrapUserMysql ) =~ /HASH/ &&
			 ref( $wrapUserMysql->{'EXECUTE'} ) =~ /CODE/ ) {

    my ($sql,$ptrRow,@accesshosts,$accesshost);

    $sql = "SELECT password,host FROM user WHERE user='$user'";

    if ( $wrapUserMysql->{'EXECUTE'}->( $sql ) ) {

			$item = makeXmlNode( 'dbuser' );
			$item->{'ATTRIBUTE'}->('name', $user);

      while ( $ptrRow = $wrapUserMysql->{'FETCHROW'}->() ) {
				unless ( ref($passwordNode) =~ /HASH/ ) {
					$passwordNode = makePasswordNode( $ptrRow->[0] );
				}
				push @accesshosts, $ptrRow->[1];
      }

			if ( $passwordNode ) {
				$item->{'ADDCHILD'}->( $passwordNode );
			}

			if (defined($accessHosts) && ref($accessHosts) =~ /ARRAY/) {
			  @accesshosts = @{$accessHosts};
			}

			foreach $accesshost ( @accesshosts ) {
				$item->{'ADDCHILD'}->( makeXmlNode( 'accesshost', $accesshost ) );
			}

			$dbNode->{'ADDCHILD'}->( $item );

    }

    $wrapUserMysql->{'FINISH'}->();
  }

	return $item;
}

my $lastDumpId = 0;

sub generateDumpId {
  return $$ . "-" . $lastDumpId++;
}

#########################################
#
#	Dumps array given by reference into
#	temporary file.
#
#########################################
sub dumpArrayToTemporaryFile( $ )
{
	my ($arrayRef) = @_;

	use POSIX;

	my $list_filename;
	for (;;) {
		$list_filename = tmpnam();
		sysopen(TMP, $list_filename, O_RDWR | O_CREAT | O_EXCL) && last;
	}

	foreach my $line (@{$arrayRef}) {
		print TMP $line . "\n";
	}
	close TMP;

	return $list_filename;
}

#################################################
#
#	Check that given directory may be archived
#
#################################################
sub checkDirForArchive( $ )
{
	my ($srcDir) = @_;

	if (!-d $srcDir) {
		return;
	}

	# check that directory is not empty
	if (!opendir(SRCDIR, $srcDir)) {
		return;
	}

	my $filename;

	while (defined ($filename = readdir SRCDIR)) {
		next if $filename =~ /^\.\.?$/;
		# directory is not empty
		closedir(SRCDIR);
		return 1;
	}

	# directory is empty
	closedir(SRCDIR);
	return;
}

#
# Compatibilty wrapper
#
# conversions:
# $ptrNames: ['.'], '.', '' -> undef, 'foobar' -> ['foobar']
# $ptrExcludes: [], '' -> undef
#

#
# sub makeDumpFile - make a dump of dir tree (use 'tar')
#
# Arguments:
#          $destFile - name of destination file.
#                      If the file has not '.tgz' or '.tar.gz' extension
#                      '.tgz' will be added
#          $scrDir   - path to source dir
#
#          $ptrNames - pointer to an array with names of files and dirs to
#                      include into dump
#                      can be a single name
#                      by default '.' - include all entries
#
#          $ptrExcludes - pointer to an array with names of files and dirs to
#                      exclude from dump
#                      can be a single name
#
#          $extraParams - extra params to run tar
#                         ('-h' - don't dump symlinks; dump the files they point to)
#

sub makeDumpFile {
  my ($destFile, $srcDir, $ptrNames, $ptrExclude, $extraParams, $user, $noRecursion) = @_;

  return unless -d $srcDir;

  if (substr($destFile, 0, length($workDir)) eq $workDir) {
    $destFile = substr($destFile, length($workDir));
  }

  if (substr($destFile, 0, length("pma/")) eq "pma/") {
    $destFile = substr($destFile, length("pma/"));
  }

  return if exists $ptrArgs->{'no-content'} or exists $ptrArgs->{'configuration-only'};

  my $followSymlinks = 0;
  if (ref($extraParams) eq "ARRAY" and @{$extraParams} > 0 and $extraParams->[0] eq '-h') {
    $followSymlinks = 1;
  }

  my $include;
  if (ref($ptrNames) eq "ARRAY") {
    if (@{$ptrNames} > 0 and !(@{$ptrNames} == 1 and $ptrNames->[0] eq '.')) {
      $include = $ptrNames;
    }
  } elsif (defined $ptrNames) {
    if ($ptrNames ne '.' and $ptrNames ne '') {
      $include = [$ptrNames];
    }
  }

  my $exclude;
  if (ref($ptrExclude) eq "ARRAY") {
    if (@{$ptrExclude} > 0) {
      $exclude = $ptrExclude;
    }
  } elsif (defined $ptrExclude) {
    if ($ptrExclude ne '') {
      $exclude = [$ptrExclude];
    }
  }

  my %options;
  $options{'directory'} = $srcDir;
  $options{'include'} = $include if $include;
  $options{'exclude'} = $exclude if $exclude;
  $options{'follow_symlinks'} = $followSymlinks if $followSymlinks;
  $options{'user'} = $user if $user;
  $options{'no_recursion'} = 1 if $noRecursion;

  return $storage->addTar($destFile, %options);
}

sub dumpFile {
  my ($destFile, $srcFile, $extraParams) = @_;
  return unless -f $srcFile;

  my @parts = split /\//, $srcFile;
  my $basename = pop @parts;
  my $dirname = join "/", @parts;

  return makeDumpFile($destFile, $dirname, $basename, undef, $extraParams);
}

sub getDumpDir {
  my ($dumpdir) = @_;

  unless(system("rm -rf $dumpdir")==0){
    &printToError("Error: can not clean temporary dir '$dumpdir': $!");
  }
  unless(-d $dumpdir){
    mkdir ($dumpdir,0755) or
      die "Error: can not create temporary dir '$dumpdir': $!\n";
  }
}

sub printXml {
  my ($root, $outFh) = @_;

  ref($root) =~ /HASH/ && ref($root->{'PRINT'}) =~ /CODE/
    or die "Error: sub printXml: bad arguments";

  print $outFh "<?xml version='1.0' encoding='UTF-8'?>\n";

  $root->{'PRINT'}->($outFh, '  ');

  if (ref($out) !~ /GLOB/) {
    close OUT;
  }
  return 1;
}

#
# XmlNode("blah", "content" => "blah",
#                 "children" => [$foo, $bar],
#                 "attributes" => {"botva"=>"7", "foo"=>"bar"})
#
sub XmlNode {
  my ($name, %args) = @_;

  my $node;
  if ($args{'content'}) {
	$node = makeXmlNode($name, $args{'content'});
  } else {
	$node = makeXmlNode($name);
  }

  if ($args{'attributes'}) {
	foreach $attribute (keys %{$args{'attributes'}}) {
	  $node->{'ATTRIBUTE'}->($attribute, $args{'attributes'}->{$attribute});
	}
  }

  if ($args{'children'}) {
	foreach $child (@{$args{'children'}}) {
	  $node->{'ADDCHILD'}->($child);
	}
  }
  return $node;
}

sub makeXmlNode {
  my ($this,$parent,%attributes,@content);
  my ($name,$plain)=@_;

  $plain = toUtf8(encodeHtml($plain));

  unless($name||$plain){
    return undef;
  }
  $this = {
	   'ATTRIBUTE' => sub {
	     my $key = shift;
	     if(@_){
	       my $value=shift;
		   die "Undefined value of attribute '$key'" unless defined($value);
	       if(@_){
					 $attributes{$key} = $value;
	       }else{
					 $attributes{$key} = toUtf8(normalizeAttribute($value));
	       }
	     }
	     return $attributes{$key};
	   },
	   'PARENT' => sub {
	     if(@_){
	       my $arg = shift;
	       if(ref($arg)=~/HASH/){
					 $parent=shift;
	       }
	     }
	     return $parent;
	   },
	   'NAME' => sub {
	     if(@_){
	       $name = shift;
	     }
	     return $name;
	   },
	   'CONTENT' => sub {
	     if(@_){
	       $plain = toUtf8(encodeHtml(shift));
	     }
	     return $plain;
	   },
	   'ADDCHILD'=> sub {
	     my $kind = shift;
	     if(
					(ref($kind)=~/HASH/) &&
					(ref($kind->{'PARENT'})=~/CODE/)
	       ) {
	       push @content,$kind;
	       $kind->{'PARENT'}->($this);
	     }
	     return $#content;
	   },
	   'KIND'=> sub {
	     my $index = shift;
			 my ($ret);
			 if ( $index =~ /^\d+$/ ) {
				 if($index >= 0 && $index <= $#content ) {
					 $ret =  $content[$index];
				 }
			 } else {
				 foreach my $item (@content){
					 if ( $item->{'NAME'}->() eq $index ) {
						 $ret = $item;
						 last;
					 }
				 }
			 }
			 return $ret;
	   },
	   'KINDSREF'=> sub {
			 my $name = shift;
			 my $ret = \@content;
			 if ( $name ) {
				 $ret = [];
				 foreach my $item (@content) {
					 if ( $item->{'NAME'}->() eq $name ) {
						 push @{$ret}, $item;
					 }
				 }
			 }
	     return $ret;
	   },
	   'ATTRIBUTESREF' => sub {
	     return \%attributes;
	   },
	   'PRINT'=>sub {
	     my $fh = shift || *STDOUT;
	     my $addPrefix = shift || '';
	     my $prefix = shift || '';
	     if($name){
	       my ($out,$item);
	       if(@content || defined($plain)){

					 $out = $prefix . &getTag($name,\%attributes,'start');
					 print $fh $out;
					 if(@content){
						 print $fh "\n";
						 foreach $item (@content){
							 if(ref($item)=~/HASH/ && ref($item->{'PRINT'})=~/CODE/){
								 $item->{'PRINT'}->($fh,$addPrefix,$prefix.$addPrefix);
							 }
						 }
					 }
					 if(defined($plain)){
						 print $fh $plain;
						 if(@content){
							 print $fh "\n$prefix";
						 }
					 }else{
						 if($prefix){
							 print $fh $prefix;
						 }
					 }
					 $out = &getTag($name,\%attributes,'stop') . "\n";
					 print $fh $out;

	       }else{
					 $out = $prefix . &getTag($name,\%attributes) . "\n";
					 print $fh $out;
	       }
	     }else{
	       if(defined($plain)){
					 print $fh $plain;
	       }
	     }
	   }
	 };

  return $this;
}

sub getTag {
  my ($name,$ptrAttr,$mode) = @_;
  my ($out,$key,$value);
  $mode||='';
  if ($mode eq 'stop'){
    $out="</$name>";
  }else{
    $out="<$name";
    while (($key,$value)=each(%{$ptrAttr})){
      $out .= ' '.$key.'="'.$value.'"';
    }
    if($mode eq 'start'){
      $out .= ">";
    }else{
      $out .= "/>";
    }
  }
  return $out;
}

#
# end class for handle with XML
#

sub getDbConnect {
  my ($ptrBin, $myDbType, $myDbAddress, $myDbUser, $myDbPasswd, $myDbDB, $myDbHost,
      $continueAfterError, $myDbSocket, $variables, $utf) = @_;

  my $connection;

  if ($myDbType eq 'Pg') {
    $myDbType = 'postgresql';
  }

  my %params;
  $params{'name'} = $myDbDB;
  $params{'user'} = $myDbUser;
  $params{'password'} = $myDbPasswd;
  $params{'host'} = $myDbHost if $myDbHost;
  $params{'socket'} = $myDbSocket if $myDbSocket;
  $params{'preload_dirs'} = $variables if $variables;
  $params{'type'} = $myDbType;

  $connection = Db::Connection::getConnection(%params);

  $this =
    {
     'DISCONNECT' => sub { return $connection->disconnect(@_); },
     'EXECUTE' => sub { return $connection->execute_rownum(@_); },
     'FETCHROW' => sub { return $connection->fetchrow(@_); },
     'FETCHHASH' => sub { return $connection->fetchhash(@_); },
     'FINISH' => sub { return $connection->finish(@_); },
     'PING' => sub { return $connection->ping(@_); },
     'USER' => sub { return $myDbUser; },
     'PASSWORD' => sub { return $myDbPasswd; },
     'HOST' => sub { return $myDbHost; }
    };
  return $this;
}

#
# end class for handle database
#

sub makeMIMEBase64 {
  my ($useModule,$this);
  if (eval "require MIME::Base64"){
    $useModule = 1;
  }else{
    $useModule = 0;
  }

  $this={
	 'ENCODE'=>sub{
	   my $src = shift;
	   my ($encoded);
	   if($useModule){
	     $encoded = MIME::Base64::encode($src);
		 chomp $encoded;
	   }else{
	     $encoded = &encodeBase64($src);
	   }
	   return $encoded;
	 },
	 'USEMODULE'=>sub{
	   return $useModule;
	 },
	};

  return $this;
}

#
# url decode (certificates in Plesk DB encoded by the url encoding)
#

sub urlDecode {
    my $url = $_[0];
    $url =~ tr/+/ /;
    $url =~ s/%([a-fA-F0-9]{2,2})/chr(hex($1))/eg;
    return $url;
}

sub encodeHtml {
  my(@ret,$src);
  if ($src=shift){
    if(ref $src){
      if(${$src}=~/[&"'><]/){
		${$src}=~s/&/&amp;/sg;
		${$src}=~s/>/&gt;/sg;
		${$src}=~s/</&lt;/sg;
		${$src}=~s/\'/&#39;/sg;
		${$src}=~s/\"/&quot;/sg;
      }
      return ${$src};
    }else{
      if($src=~/[&"'><]/){
		$src=~s/&/&amp;/sg;
		$src=~s/>/&gt;/sg;
		$src=~s/</&lt;/sg;
		$src=~s/\'/&#39;/sg;
		$src=~s/\"/&quot;/sg;
      }
    }
  }
  return $src;
}

# Normalization according to the XML 1.0 (Third Edition),
# clauses 2.3 "Common Syntactic Constructs"
# and 3.3.3 "Attribute-value normalization"

sub normalizeAttribute {
  $_ = $_[0];
  s/\&/&amp;/g;
  s/\r/&#xD;/g;
  s/\n/&#xA;/g;
  s/\t/&#x9;/g;
  s/\</&lt;/g;
  s/\"/&quot;/g;
  s/\'/&#39;/g;
  return $_;
}

############################
#
#	Non-ascii characters handling
{
	# static
	my $charset;

	sub setCharset {
	  $charset = shift;
	}

	# try to determine the local charmap
	sub _determineCharset()
	{
	  my $charset;
	  my $locale = $bin{'locale'};

	  if (exists $ENV{'LANG'}) {
		my @pieces = split /\./, $ENV{'LANG'};
		if (defined $pieces[1]) {
		  $charset = $pieces[1];
		  chomp $charset;
		}
	  }
	  if (not $charset and $locale and open(CM, "$locale charmap |")) {
	  	binmode CM;
		$charset = <CM>;
		close(CM);
		chomp $charset;
	  }

      $charset = 'ISO-8859-1' unless $charset;

	  return $charset;
	}

	sub iconv {
	  if (open(IC, "echo '$_[0]' | " . $bin{'iconv'} . " -f '$_[1]' -t 'UTF-8' |")) {
		binmode IC;
		my $converted = join "", <IC>;
		chomp $converted;
		close IC;
		return $converted if $converted;
	  }
	}

	sub toUtf8 {
		my ($subj, $myCharset) = @_;

		$charset = _determineCharset() if (not $charset);

		$myCharset = $charset unless ( $myCharset );

		if (!$subj || $subj =~ /^[\x00-\x7f]*$/) {
		  return $subj;
		}

		if ($myCharset =~ /UTF-?8/i) {
          if (Utf8Checker::check($subj)) {
            return $subj;
          } else {
			printToLog("UTF-8 charset detected, but invalid UTF-8 sequence occured");
			return iconv($subj, "ISO-8859-1");
		  }
		}

		$subj =~ s/\'/\'"\'"\'/g;
		return iconv($subj, $myCharset);
	}
}

#
#   Converts localized domain name into IDN ASCII representation
#
sub utf8ToIdn {
    my ($utf8domain) = @_;

    unless (AgentConfig::idnconvBin()) {
        return $utf8domain;
    }

    if (open(IC, "echo '$utf8domain' | " . AgentConfig::idnconvBin() . " -in 'UTF-8' |")) {
        binmode IC;
        my $converted = join "", <IC>;
        chomp $converted;
        close IC;
        return $converted if $converted;
    }

    return $utf8domain;
}

#
#   Converts IDN ASCII domain representation into UTF8 localized name
#
sub idnToUtf8 {
    my ($idnDomain) = @_;

    unless (AgentConfig::idnconvBin()) {
        return $idnDomain;
    }

    if (open(IC, "echo '$idnDomain' | " . AgentConfig::idnconvBin() . " -reverse -out 'UTF-8' |")) {
        binmode IC;
        my $converted = join "", <IC>;
        chomp $converted;
        close IC;
        return $converted if $converted;
    }

    return $idnDomain;
}

#
# Based on the b64.c by Bob Trower (base64.sf.net)
#

my @cb64 = ('A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
			'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
			'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
			'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
			'g', 'h', 'i', 'j',	'k', 'l', 'm', 'n',
			'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
			'w', 'x', 'y', 'z', '0', '1', '2', '3',
			'4', '5', '6', '7', '8', '9', '+', '/');

@cb64 = map { ord } @cb64;

sub encodeBlock {
  my ($len, @in) = @_;
  my @res;
  push @res, $cb64[$in[0] >> 2];
  push @res, $cb64[(($in[0] & 0x3)<<4)|(($in[1]&0xf0)>>4)];
  push @res, $len>1 ? $cb64[(($in[1]&0xf)<<2)|(($in[2]&0xc0)>>6)] : ord('=');
  push @res, $len>2 ? $cb64[$in[2]&0x3f] : ord('=');
  return @res;
}

sub encodeBase64 {
  my @in = unpack "C*", $_[0];

  my @out;
  while (scalar(@in)) {
	my @block = (0, 0, 0);
	my $length = 0;
	while (scalar(@in) && $length < 3) {
	  $block[$length++] = shift(@in);
	}
	push @out, encodeBlock($length, @block);
  }
  return pack "C*", @out;
}

sub resolveFile {
  my ($prefixes, $file) = @_;

  foreach my $prefix (@{$prefixes}) {
	$prefix.='/' unless $prefix=~/\/$/;
	$file =~ s|^/||;

	return ($prefix.$file) if (-x ($prefix.$file));
  }
}

sub checkGnuTar {
  my ($file) = @_;

  # on some FreeBSD this command outputs on stderr
  open (TAR, "$file --version 2>&1 |");
  while (<TAR>) {
    if (/GNU tar/) {
      close TAR;
      return 1;
    }
  }
  close TAR;
  return 0;
}

sub getBins {
  my ($ptrBin,$path) = @_;
  my (%searchPath,$key,$prefix,$bin,$fullPath);
  my ($notFound,$i,@prefixes);

  $notFound=0;

  $i=0;
  foreach $prefix (split(/:/,$path)){
    next unless $prefix;
    unless(exists($searchPath{$prefix})){
      $searchPath{$prefix}=$i++;
      push @prefixes,$prefix;
    }
  }

  foreach $key (keys %{$ptrBin}){
	$ptrBin->{$key} = resolveFile(\@prefixes, $ptrBin->{$key});

	# special case: 'tar' *must* be GNU tar

	if ($key eq 'tar') {
	  my $foundGnuTar = checkGnuTar($ptrBin->{'tar'});

	  # if 'tar' is not GNU tar, try 'gtar'
	  if (!$foundGnuTar) {
		if (checkGnuTar(resolveFile(\@prefixes, 'gtar'))) {
		  $ptrBin->{'tar'} = resolveFile(\@prefixes, 'gtar');
		} else {
		  $ptrBin->{'tar'} = undef;
		}
	  }

	  if ($ptrBin->{'tar'}) {
		open (TAR, "$ptrBin->{'tar'} --version|");
		while (<TAR>) {
		  if (/tar \(GNU tar\) (\d+)\.(\d+)/) {
			if ($1 > 1 or ($1 == 1 and $2 >= 13)) {
			  $ptrBin->{'tar-accepts-doubledash'} = 1;
			}
			last;
		  }
		}
		close (TAR);
	  }
	}

	# bsdtar case: it is fallback case and should be set up
	# 1) to 'tar' if 'tar' is not GNU tar
	# 2) to 'bsdtar' if no 'tar' found in system

	if ($key eq 'bsdtar') {
	  # if 'bsdtar' not found, but 'tar' exists and not a GNU tar
	  unless ($ptrBin->{'bsdtar'}) {
		if (resolveFile(\@prefixes, 'tar')) {
		  unless (checkGnuTar(resolveFile(\@prefixes, 'tar'))) {
			$ptrBin->{'bsdtar'} = resolveFile(\@prefixes, 'bsdtar');
		  }
		}
	  }
	}

    $notFound++;
  }

  return $notFound;
}

sub getArguments {
  my ($ptrArgv,$ptrOpts) = @_;

  my (@keys,$firstKey,%prepKeys,$key,$value,$ptrArr,$arg,$state);
  my (%retHash,$pat,$found,$used,@rest,$fullArg,$prevKey);

  while (($key,$value)=each(%{$ptrOpts})){
    @keys = split(/\|/,$key);
    $firstKey = $keys[0];
    $firstKey =~s/^-*//;
    push @{$prepKeys{$firstKey}},[@keys];
    push @{$prepKeys{$firstKey}},$value;
  }

  $state =0;
  foreach $arg (@{$ptrArgv}){

    $used = 0;
    if($state==1){
      if  ($arg =~ /^-/){
	$state=0;
      }else{
	$used = 1;
	$retHash{$prevKey}=$arg;
	next;
      }
    }
    if ($state == 2){
      $retHash{$prevKey}=$arg;
      $state = 0;
      $used = 1;
    }else{
      $fullArg = $arg;
      if ($arg =~ /^(-\S+)=(.+)/s){
	$arg = $1;
	$value = $2;
      }else{
	$value = undef;
      }
      foreach $key (keys %prepKeys){

	$ptrArr = $prepKeys{$key};

#	print "key: $key, ",join('|',@{$ptrArr->[0]}),"\n";


	$found = 0;
	foreach $pat (@{$ptrArr->[0]}){
	  if ($pat eq $arg){
	    $found = 1;
	    last;
	  }
	}
	if($found){
	  $used = 1;
	  if(defined($value)){
	    $retHash{$key}=$value;
	  }else{
	    if($ptrArr->[1]){
	      if($ptrArr->[1] =~ /\?$/){
		$state = 1;
	      }else{
		$state = 2;
	      }
	      $prevKey = $key;
	    }else{
	      $retHash{$key}=undef;
	    }
	  }
	  last;
	}
      }
    }
    unless($used){
      if($value){
	push @rest,$fullArg;
      }else{
	push @rest,$arg;
      }
    }
  }
  @{$ptrArgv}=@rest;
  return \%retHash;

}

sub makeDumpStatus($) {
  my $statusFile = shift;
  my($pid,$currAccount,$currDomain,$leftAccounts,$leftDomains,$this);
  $pid=$$;
  $this = {'FILE'=>sub {
	     if(@_){
	       $statusFile = shift;
	     }
	     return $statusFile;
	   },
	   'PID'=>sub {
	     if(@_){
	       $pid = shift;
	     }
	     return $pid;
	   },
	   'ACCOUNT'=>sub{
	     if(@_){
	       $currAccount = shift;
	       if($leftAccounts){
		 $leftAccounts--;
	       }
	     }
	     return $currAccount;
	   },
	   'DOMAIN'=>sub{
	     if(@_){
	       $currDomain = shift;
	       if($leftDomains){
		 $leftDomains--;
	       }
	     }
	     return $currDomain;
	   },
	   'COUNTS'=>sub{
	     if(@_){
	       my $newValue = shift;
	       if (defined $newValue){
		 $leftAccounts=$newValue;
	       }
	       $newValue = shift;
	       if (defined $newValue){
		 $leftDomains=$newValue;
	       }
	     }
	     return ($leftAccounts,$leftDomains);
	   },
	   'PRINT'=>sub{
	     if(@_){
	       $statusFile=shift;
	     }
	     open(STATUS,">$statusFile.tmp");

		 binmode STATUS;
	     my $root = &makeXmlNode('agent-dump-status');
	     my $item = &makeXmlNode('pid',$pid);
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('account',$currAccount||'');
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('domain',$currDomain||'');
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('accounts-left',
				  defined($leftAccounts)?$leftAccounts:'');
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('domains-left',
				  defined($leftDomains)?$leftDomains:'');
	     $root->{'ADDCHILD'}->($item);

	     &printXml($root,*STATUS,'agent-dump-status.dtd');

	     close(STATUS);

		 rename("$statusFile.tmp", $statusFile);
	   }

	  };
  return $this;
}

sub makeContentList {
  #FIXME: HACK
  my $workDir = `pwd`;
  chomp $workDir;
  open CONTENT_LIST, "< $workDir/content-list.xml";
  print while(<CONTENT_LIST>);
  close CONTENT_LIST;
}

my $outputIsFile;

sub openOutput
{
	my ($outPath) = @_;
	my ($outFh);

	if ($outPath) {
		open(OUTPUT, ">>$outPath") or die "Error: can not open '$outPath' for write: $!\n";
		binmode OUTPUT;
		$outFh = *OUTPUT;
		$outputIsFile = 1;
	} else {
		$outFh = *STDOUT;
		$outputIsFile = 0;
	}

	return $outFh;
}

sub closeOutput{
  if($outputIsFile){
    close (OUTPUT) or
      &printToError("Error: unable to close OUTPUT: $!\n");
    $outputIsFile=0;
  }
}

#=========================================================================
#
# dtd-specific (but common) subroutines
#
sub createPasswordNode {
  my ($parent, $password, $type) = @_;

  my $item = makeXmlNode('password', $password);
  $item->{'ATTRIBUTE'}->('type', $type);
  $parent->{'ADDCHILD'}->($item);
}

sub createSysuserNode {
  my ($parent, $name, $password, $type) = @_;

  my $item = makeXmlNode('sysuser');
  $item->{'ATTRIBUTE'}->('name', $name);
  createPasswordNode($item, $password, $type);
  $parent->{'ADDCHILD'}->($item);
}

sub addPermissionNode {
	my ($parent, $name, $allowed) = @_;

	my $item = makeXmlNode('permission');
	$item->{'ATTRIBUTE'}->('name', $name);
	$item->{'ATTRIBUTE'}->('allowed', $allowed);
	$parent->{'ADDCHILD'}->($item);
}

sub addMailuserPermissionNode {
	my ($parent, $name, $allowed) = @_;

	my $item = makeXmlNode('mailuser-permission');
	$item->{'ATTRIBUTE'}->('name', $name);
	$item->{'ATTRIBUTE'}->('allowed', $allowed);
	$parent->{'ADDCHILD'}->($item);
}

sub readFile {
  my ($fileName) = @_;

  unless (-f $fileName) {
	return undef;
  }

  open(INPUT, "< $fileName");
  binmode INPUT;
  my $content = join('', <INPUT>);
  close(INPUT);

  return $content;
}

# Returns array with removed duplicates
sub uniq( $ )
{
	my %seen = ();
	return grep { ! $seen{$_} ++ } @_;
}

# Returns string, suitable for using in 'IN ($list)' sql clause
sub getSqlList( $ )
{
	return join(',', map {"'$_'"} sort {$a <=> $b} @_);
}

sub getMysqlVersion {
  my ($ptrHash) = @_;
  my %bin=%{$ptrHash};
  my $mysql = $bin{'mysql'};
  my $version = `$mysql --version`;
  if ($version and $version =~ /Distrib (\d+\.\d+)\./){
    return $1;
  }
}

# Returns random password
sub randomPasswd() {
	my $symbols = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789';
	my $seq_length = length($symbols);
	my $passwd = '';

	for ($i = 0; $i < 7; ++$i) {
		$passwd .= substr($symbols, rand($seq_length), 1);
	}

	return $passwd;
}


1;

#
# end common subroutines
#
#=========================================================================
